"""Top-level package for cfglib.

Objects from the `config` and `spec` submodules are reexported to the root, so you can use
e.g. `cfglib.Config`.
"""

__author__ = '''Anton Barkovsky'''
__email__ = 'anton@swarmer.me'
__version__ = '1.0.1'


# pylint: disable=wildcard-import
from .config import *
from .spec import *



from collections import abc as collections_abc
from typing import *

from ..config import ConfigProjection, ProjectedConfig, ProxyConfig
from ..spec import MISSING


# pylint: disable=too-many-ancestors
class ArgsNamespaceConfig(ProjectedConfig):
    """A config that takes its contents from the environment."""

    def __init__(
        self,
        args: object,
        uppercase: bool = False,
        relevant_keys: Optional[Iterable[str]] = None,
    ):
        projection = ArgsNamespaceConfigProjection(uppercase, relevant_keys)
        args_data = {k: v for k, v in args.__dict__.items() if v is not MISSING}
        super().__init__(ProxyConfig(args_data), projection)


class _Universe(collections_abc.Container):
    def __contains__(self, item):
        return True


class ArgsNamespaceConfigProjection(ConfigProjection):
    """A commonly used projection for cmdline args variables:
    filter and optionally uppercase keys.
    """

    def __init__(
        self,
        uppercase: bool = False,
        relevant_keys: Optional[Iterable[str]] = None,
    ):
        self.uppercase = uppercase
        self.relevant_keys: Container = (
            frozenset(relevant_keys)
            if relevant_keys is not None
            else _Universe()
        )

    def is_relevant_key(self, key: str) -> bool:
        if self.uppercase and not key.isupper():
            return False

        return self.is_relevant_sourcekey(self.key_to_sourcekey(key))

    def is_relevant_sourcekey(self, sourcekey: str) -> bool:
        if self.uppercase and not sourcekey.islower():
            return False

        return sourcekey in self.relevant_keys

    def key_to_sourcekey(self, key: str) -> str:
        if self.uppercase:
            key = key.lower()

        return key

    def sourcekey_to_key(self, sourcekey: str) -> str:
        if self.uppercase:
            sourcekey = sourcekey.upper()

        return sourcekey



# pylint: disable=too-many-ancestors
from __future__ import annotations

import abc
import collections.abc
from itertools import chain
from typing import *


__all__ = [
    'Config',
    'MutableConfig',
    'DictConfig',
    'ProxyConfig',
    'CachingConfig',
    'CompositeConfig',
    'ConfigProjection',
    'BasicConfigProjection',
    'LOWERCASE_PROJECTION',
    'UPPERCASE_PROJECTION',
    'ProjectedConfig',
    'to_cfg',
    'to_cfg_list',
]


def to_cfg(value: Any) -> Config:
    if isinstance(value, Config):
        return value
    elif isinstance(value, Mapping):
        return ProxyConfig(value)
    else:
        raise TypeError(f'Cannot convert to config: {value}')


def to_cfg_list(value: Any) -> List[Config]:
    if not isinstance(value, collections.abc.Collection) or isinstance(value, Mapping):
        value = [value]

    return [to_cfg(item) for item in value]


class Config(collections.abc.Mapping):
    """An abstract configuration interface

    A config is really just a dict with some extra methods.

    In a custom implementation you need to define at least:\n
    - __getitem__(item)\n
    - __iter__()\n
    - __len__()\n
    - reload()\n
    """

    def snapshot(self) -> 'DictConfig':
        """Return a copied snapshot of this config, backed by memory
        """
        return DictConfig(self)

    @abc.abstractmethod
    def reload(self):
        """Reload all config items from its backing store. The contents may change arbitrarily.

        Not all configs can meaningfully reload. In this case this method will do nothing.
        Some configs always pull fresh data. In that case this method also does nothing.
        """
        pass  # pragma: no cover


class MutableConfig(abc.ABC, collections.abc.MutableMapping, Config):
    """An abstract base class for mutable configs (with __setitem__)"""
    pass


class DictConfig(dict, MutableConfig):
    """A config backed by its own dictionary stored in memory. In other words, a fancy dict."""

    def reload(self):
        """Does nothing."""
        pass

    def replace(self, other: Config):
        """Update self in place to become a shallow copy of *other*"""
        self.clear()
        self.update(other)


class ProxyConfig(MutableConfig):
    """A config that uses a separate mapping as source.

    Unlike DictConfig, this is not by itself a dict, it only references a dict.

    Althouth this class has basically zero logic, it's useful to:
    - Wrap any mapping to conform to the Config interface
    - Be a level of indirection to switch source configs easily by changing .source field.
    """

    def __init__(self, source: Mapping):
        self.source = source
        self._mutable = isinstance(source, collections.abc.MutableMapping)

    def __getitem__(self, key):
        return self.source[key]

    def __setitem__(self, key, value):
        if not self._mutable:
            raise TypeError('ProxyConfig\'s source is not mutable')

        self.source[key] = value

    def __delitem__(self, key):
        if not self._mutable:
            raise TypeError('ProxyConfig\'s source is not mutable')

        del self.source[key]

    def __len__(self):
        return len(self.source)

    def __iter__(self):
        return iter(self.source)

    def reload(self):
        """Reload source if it's a Config, otherwise do nothing."""
        if isinstance(self.source, Config):
            self.source.reload()


class CachingConfig(DictConfig):
    """A config that copies data from a wrapped config once and returns data from this copy,
    until manually reloaded."""

    def __init__(self, wrapped_config: Config):  # type: ignore
        super().__init__()

        self.wrapped_config = wrapped_config
        self.replace(self.wrapped_config)

    def reload(self):
        """Refresh the underlying config and update the cache."""
        self.wrapped_config.reload()
        self.replace(self.wrapped_config)


class CompositeConfig(Config):
    """A config backed by multiple configs

    Entries are searched in subconfigs in reverse order and the first found
    value is returned. That is, the first subconfig has the lowest priority,
    and the last subconfig takes precedence over all others.
    """

    def __init__(self, subconfigs: Iterable[Config]):  # type: ignore
        self.subconfigs = list(subconfigs)

    def __getitem__(self, item):
        for subconfig in self.subconfigs[::-1]:
            try:
                return subconfig[item]
            except KeyError:
                continue

        raise KeyError(f'Key {item} not found in any subconfig')

    def __iter__(self):
        return iter(self._all_keys)

    def __len__(self):
        return len(self._all_keys)

    def __repr__(self):
        snapshot = self.snapshot()
        return f'<CompositeConfig {snapshot}>'

    @property
    def _all_keys(self) -> frozenset:
        all_keys = frozenset(chain.from_iterable(
            iter(subconfig)
            for subconfig in self.subconfigs
        ))
        return all_keys

    def reload(self):
        """Reload subconfigs."""
        for subconfig in self.subconfigs:
            subconfig.reload()


class ConfigProjection(abc.ABC):  # pragma: no cover
    """ABC for a projection to be passed to a `ProjectedConfig`.

    The projection should be consistent, that is:\n
    - k2sk(sk2k(sk)) == sk for all sk that are relevant\n
    - sk2k(k2sk(k)) == k for all k that are relevant\n
    - is_relevant(key) <=> is_relevant_sourcekey(k2sk(k))
    """

    @abc.abstractmethod
    def is_relevant_key(self, key: str) -> bool:
        """Should a projected config key be accepted by the ProjectedConfig?"""
        ...

    @abc.abstractmethod
    def is_relevant_sourcekey(self, sourcekey: str) -> bool:
        """Should a source config key be used by the ProjectedConfig?"""
        ...

    @abc.abstractmethod
    def key_to_sourcekey(self, key: str) -> str:
        """Map a ProjectedConfig's key to a source config's key"""
        ...

    @abc.abstractmethod
    def sourcekey_to_key(self, sourcekey: str) -> str:
        """Map source config's key to a a ProjectedConfig's key"""
        ...


class BasicConfigProjection(ConfigProjection):
    """A helper to easily create a ConfigProjection"""

    def __init__(
        self,
        is_relevant_key: Optional[Callable] = None,
        is_relevant_sourcekey: Optional[Callable] = None,
        key_to_sourcekey: Optional[Callable] = None,
        sourcekey_to_key: Optional[Callable] = None
    ):
        if is_relevant_key:
            self._is_relevant_key = is_relevant_key
        else:
            def _default_is_relevant_key(key):
                return self.is_relevant_sourcekey(self.key_to_sourcekey(key))

            self._is_relevant_key = _default_is_relevant_key

        if is_relevant_sourcekey:
            self._is_relevant_sourcekey = is_relevant_sourcekey
        else:
            def _default_is_relevant_sourcekey(_key):
                return True

            self._is_relevant_sourcekey = _default_is_relevant_sourcekey

        if key_to_sourcekey:
            self.key_to_sourcekey = key_to_sourcekey  # type: ignore

        if sourcekey_to_key:
            self.sourcekey_to_key = sourcekey_to_key  # type: ignore

    # pylint: disable=method-hidden
    def is_relevant_key(self, key: str) -> bool:
        """Check that sk2k(k2sk(k)) maps key to itself and, by default,
        compute source key and defer to is_relevant_sourcekey."""
        if self.sourcekey_to_key(self.key_to_sourcekey(key)) != key:
            return False

        return self._is_relevant_key(key)

    # pylint: disable=method-hidden
    def is_relevant_sourcekey(self, sourcekey: str) -> bool:
        """Check that k2sk(sk2k(sk)) maps sourcekey to itself and, by default, return True."""
        if self.key_to_sourcekey(self.sourcekey_to_key(sourcekey)) != sourcekey:
            return False

        return self._is_relevant_sourcekey(sourcekey)

    # pylint: disable=method-hidden
    def key_to_sourcekey(self, key: str) -> str:
        """By default, identity function."""
        return key

    # pylint: disable=method-hidden
    def sourcekey_to_key(self, sourcekey: str) -> str:
        """By default, identity function."""
        return sourcekey


LOWERCASE_PROJECTION = BasicConfigProjection(
    key_to_sourcekey=lambda k: k.upper(),
    sourcekey_to_key=lambda sk: sk.lower(),
)

UPPERCASE_PROJECTION = BasicConfigProjection(
    key_to_sourcekey=lambda k: k.lower(),
    sourcekey_to_key=lambda sk: sk.upper(),
)


class ProjectedConfig(MutableConfig):
    """Config that renames or filters source config's keys."""

    def __init__(self, subconfig: Config, projection: ConfigProjection):
        self.subconfig = subconfig
        self.projection = projection

    def __getitem__(self, key):
        if not self.projection.is_relevant_key(key):
            raise KeyError(f'Key {key} not relevant')

        sourcekey = self.projection.key_to_sourcekey(key)
        return self.subconfig[sourcekey]

    def __setitem__(self, key, value):
        if not isinstance(self.subconfig, MutableConfig):
            raise TypeError('ProjectedConfig\'s subconfig is not mutable')

        if not self.projection.is_relevant_key(key):
            raise KeyError(f'Key {key} not relevant')

        sourcekey = self.projection.key_to_sourcekey(key)
        self.subconfig[sourcekey] = value

    def __delitem__(self, key):
        if not isinstance(self.subconfig, MutableConfig):
            raise TypeError('ProjectedConfig\'s subconfig is not mutable')

        if not self.projection.is_relevant_key(key):
            raise KeyError(f'Key {key} not relevant')

        sourcekey = self.projection.key_to_sourcekey(key)
        del self.subconfig[sourcekey]

    def __len__(self):
        return len(list(self._relevant_sourcekeys))

    def __iter__(self):
        return (
            self.projection.sourcekey_to_key(sourcekey)
            for sourcekey in self._relevant_sourcekeys
        )

    @property
    def _relevant_sourcekeys(self) -> Iterable[str]:
        return (
            sourcekey
            for sourcekey in iter(self.subconfig)
            if self.projection.is_relevant_sourcekey(sourcekey)
        )

    def reload(self):
        """Reload the source config."""
        self.subconfig.reload()



import os

from ..config import ConfigProjection, ProjectedConfig, ProxyConfig


# pylint: disable=too-many-ancestors
class EnvConfig(ProjectedConfig):
    """A config that takes its contents from the environment."""

    def __init__(self, prefix: str = '', lowercase: bool = False):
        projection = EnvConfigProjection(prefix, lowercase)

        super().__init__(ProxyConfig(os.environ), projection)


class EnvConfigProjection(ConfigProjection):
    """A commonly used projection for env variables: remove prefix and optionally lowercase keys."""

    def __init__(self, prefix: str, lowercase: bool = False):
        self.prefix = prefix
        self.lowercase = lowercase

    def is_relevant_key(self, key: str) -> bool:
        if self.lowercase and not key.islower():
            return False

        return True

    def is_relevant_sourcekey(self, sourcekey: str) -> bool:
        if self.lowercase and not sourcekey.isupper():
            return False

        return sourcekey.startswith(self.prefix)

    def key_to_sourcekey(self, key: str) -> str:
        if self.lowercase:
            key = key.upper()

        return self.prefix + key

    def sourcekey_to_key(self, sourcekey: str) -> str:
        assert sourcekey.startswith(self.prefix)
        key = sourcekey[len(self.prefix):]

        if self.lowercase:
            key = key.lower()

        return key



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['cfglib', 'cfglib.sources']

package_data = \
{'': ['*']}

extras_require = \
{'docs': ['Sphinx>=1.8,<2.0',
          'sphinxcontrib-apidoc>=0.3.0,<0.4.0',
          'sphinxcontrib-fulltoc>=1.2,<2.0']}

setup_kwargs = {
    'name': 'cfglib',
    'version': '1.0.1',
    'description': 'An extensible configuration library',
    'long_description': '======\ncfglib\n======\n\n\n.. image:: https://img.shields.io/pypi/v/cfglib.svg\n        :target: https://pypi.python.org/pypi/cfglib\n\n.. image:: https://github.com/swarmer/cfglib-py/workflows/CI/badge.svg\n        :target: https://github.com/swarmer/cfglib-py/actions\n        :alt: CI Status\n\n.. image:: https://readthedocs.org/projects/cfglib/badge/?version=latest\n        :target: https://cfglib.readthedocs.io/en/latest/?badge=latest\n        :alt: Documentation Status\n\n.. image:: https://codecov.io/gh/swarmer/cfglib-py/branch/master/graph/badge.svg\n  :target: https://codecov.io/gh/swarmer/cfglib-py\n\n\nAn extensible configuration library\n\n\n* Free software: MIT license\n* Documentation: https://cfglib.readthedocs.io.\n\n\nFeatures\n--------\n\n* Describe config schema and perform validation\n* Numerous utils to compose configs from multiple sources, etc.\n* Read settings from environment\n',
    'author': 'Anton Barkovsky',
    'author_email': 'anton@swarmer.me',
    'maintainer': None,
    'maintainer_email': None,
    'url': 'https://github.com/swarmer/cfglib-py',
    'packages': packages,
    'package_data': package_data,
    'extras_require': extras_require,
    'python_requires': '>=3.7,<4',
}


setup(**setup_kwargs)



# pylint: disable=empty-docstring
from __future__ import annotations

import enum
from typing import *

from .config import CompositeConfig, Config, DictConfig, to_cfg_list


__all__ = [
    'MISSING',

    'MissingSettingAction',
    'ERROR', 'USE_DEFAULT', 'LEAVE',

    'ValidationError',

    'Setting',
    'StringSetting',
    'BoolSetting',
    'IntSetting',
    'FloatSetting',
    'DictSetting',
    'ListSetting',

    'ConfigSpec',
    'SpecValidatedConfig',
]


class Marker:
    def __repr__(self):
        if self is MISSING:
            return 'MISSING'
        else:
            return super().__repr__()  # pragma: no cover


# Markers
MISSING = Marker()
"""A singleton marker object denoting that a setting value is (or should be) absent.
Absence here means not being in the config at all (raising KeyError on access)."""


T = TypeVar('T')  # pylint: disable=invalid-name
ExtOptional = Union[Marker, None, T]


class MissingSettingAction(enum.Enum):
    """Action to do when a setting is missing"""

    ERROR = enum.auto()
    """Raise an error"""

    USE_DEFAULT = enum.auto()
    """Set it to the provided default"""

    LEAVE = enum.auto()
    """Leave as it is (missing or None)"""


ERROR = MissingSettingAction.ERROR
USE_DEFAULT = MissingSettingAction.USE_DEFAULT
LEAVE = MissingSettingAction.LEAVE


# Exceptions
class ValidationError(Exception):
    """Raised when a setting value is not valid according to the setting parameters"""
    pass


# Setting types
class Setting:
    """Specification for one config's setting.

    :param name: Name of the setting.
        Only needs to be specified if passed to a ConfigSpec instead of a SpecValidatedConfig.
    :param default: The default value to be used when source config doesn't provide
        this setting. If the default is needed but was not specified, an error will be raised.
    :param on_missing: Action to do if this setting is entirely absent from source configs.
    :param on_null: Action to do if this setting is None in the source configs.
    """

    def __init__(
        self,
        *,
        name: Optional[str] = None,
        default: ExtOptional[Any] = MISSING,
        on_missing: MissingSettingAction = MissingSettingAction.USE_DEFAULT,
        on_null: MissingSettingAction = MissingSettingAction.LEAVE,
    ):
        self.name = name
        self.default = default
        self.on_missing = on_missing
        self.on_null = on_null

    def __set_name__(self, owner, name):
        if self.name is not None and self.name != name:
            raise ValueError(f'Setting name already set to {self.name}')

        self.name = name

    def __get__(self, instance, owner):
        if instance is None:
            return self

        return instance[self.name]

    def validate_value(self, value: Any) -> Any:
        """Validate a value and return the result (with default possibly replacing null values)"""

        if value is MISSING:
            if self.on_missing is ERROR:
                raise ValidationError(f'Config field {self.name} missing')
            elif self.on_missing is USE_DEFAULT:
                if self.default is MISSING:
                    raise ValidationError(
                        f'Config field {self.name} missing and no default is provided'
                    )

                return self.default
            elif self.on_missing is LEAVE:
                return MISSING
            else:
                raise ValueError(f'Invalid on_missing choice in field {self.name}')
        elif value is None:
            if self.on_null is ERROR:
                raise ValidationError(f'Config field {self.name} must not be None')
            elif self.on_null is USE_DEFAULT:
                if self.default is MISSING:
                    raise ValidationError(
                        f'Config field {self.name} is None and no default is provided'
                    )

                return self.default
            elif self.on_null is LEAVE:
                return None
            else:
                raise ValueError(f'Invalid on_null choice in field {self.name}')

        return self.validate_value_custom(value)

    def validate_value_custom(self, value: Any) -> Any:
        """Override this method to implement custom setting type-specific validation logic."""
        return value


class StringSetting(Setting):
    def __init__(
        self,
        *,
        default: ExtOptional[str] = MISSING,
        **kwargs,
    ):
        super().__init__(default=default, **kwargs)

    def validate_value_custom(self, value: Any) -> Optional[str]:
        """"""  # Remove the parents docstring about overriding
        if not isinstance(value, str):
            raise ValidationError(
                f'A value for setting {self.name} must be a string or None'
            )

        return value


class BoolSetting(Setting):
    def __init__(
        self,
        *,
        default: ExtOptional[bool] = MISSING,
        **kwargs,
    ):
        super().__init__(default=default, **kwargs)

    def validate_value_custom(self, value: Any) -> Optional[bool]:
        """"""  # Remove the parents docstring about overriding
        if not isinstance(value, bool):
            raise ValidationError(f'A value for setting {self.name} must be a bool')

        return value


class IntSetting(Setting):
    def __init__(
        self,
        *,
        default: ExtOptional[int] = MISSING,
        **kwargs,
    ):
        super().__init__(default=default, **kwargs)

    def validate_value_custom(self, value: Any) -> Optional[int]:
        """"""  # Remove the parents docstring about overriding
        if not isinstance(value, int):
            raise ValidationError(f'A value for setting {self.name} must be an int')

        return value


class FloatSetting(Setting):
    def __init__(
        self,
        *,
        default: ExtOptional[float] = MISSING,
        **kwargs,
    ):
        super().__init__(default=default, **kwargs)

    def validate_value_custom(self, value: Any) -> Optional[float]:
        """"""  # Remove the parents docstring about overriding
        if not isinstance(value, float):
            raise ValidationError(f'A value for setting {self.name} must be a float')

        return value


class DictSetting(Setting):
    def __init__(
        self,
        *,
        default: ExtOptional[Mapping] = MISSING,
        subtype: Union[ConfigSpec, Type[SpecValidatedConfig], None] = None,
        **kwargs,
    ):
        super().__init__(default=default, **kwargs)

        self.subtype = subtype

    def validate_value_custom(self, value: Any) -> Optional[Mapping]:
        """"""  # Remove the parent's docstring about overriding
        if not isinstance(value, Mapping):
            raise ValidationError(f'A value for setting {self.name} must be a mapping')

        if isinstance(self.subtype, ConfigSpec):
            value = self.subtype.validate_config(DictConfig(value))
        elif isinstance(self.subtype, type) and issubclass(self.subtype, SpecValidatedConfig):
            value = self.subtype([DictConfig(value)])
        elif self.subtype is None:
            pass
        else:
            raise ValueError(f'Invalid type in field {self.name}')

        return value


class ListSetting(Setting):
    def __init__(
        self,
        *,
        default: ExtOptional[List[Any]] = MISSING,
        on_empty: MissingSettingAction = MissingSettingAction.LEAVE,
        subsetting: Optional[Setting] = None,
        **kwargs,
    ):
        super().__init__(default=default, **kwargs)

        self.on_empty = on_empty
        self.subsetting = subsetting

    # noinspection DuplicatedCode
    def validate_value_custom(self, value: Any) -> ExtOptional[List[Any]]:
        """"""  # Remove the parents docstring about overriding
        if not isinstance(value, list):
            raise ValidationError(
                f'A value for a setting {self.name} must be a list or None'
            )

        if not value:
            if self.on_empty is ERROR:
                raise ValidationError(f'Config field {self.name} is empty')
            elif self.on_empty is USE_DEFAULT:
                if self.default is MISSING:
                    raise ValidationError(
                        f'Config field {self.name} empty and no default is provided'
                    )

                return self.default
            elif self.on_empty is LEAVE:
                return value
            else:
                raise ValueError(f'Invalid on_empty choice in field {self.name}')

        if self.subsetting:
            value = [
                self.subsetting.validate_value(item)
                for item in value
            ]

        return value


# Spec
class ConfigSpec:
    """A set of settings specifying some config."""

    def __init__(
        self,
        settings_iterable: Iterable[Setting],
        allow_extra: bool = False,
    ):
        settings_list = list(settings_iterable)
        self.settings: Dict[str, Setting] = {}
        for setting in settings_list:
            if setting.name is None:
                raise ValueError(f'Setting name not set for {setting}')

            self.settings[setting.name] = setting

        if len(self.settings) != len(settings_list):
            raise ValueError('All settings must have unique names')

        self.allow_extra = allow_extra

    def validate_setting(self, config: Config, setting_name: str):
        """Validate one setting of a config."""

        try:
            setting = self.settings[setting_name]
        except KeyError as exc:
            raise KeyError(f'Unknown setting, not in config spec: {setting_name}') from exc

        try:
            value = config[setting_name]
        except KeyError as exc:
            value = MISSING

        return setting.validate_value(value)

    def validate_config(self, config: Config):
        """Validate all settings of a config."""

        extra_fields = frozenset(config) - frozenset(self.settings)
        if extra_fields and not self.allow_extra:
            raise ValidationError(
                f'Unexpected fields in the config: '
                f'{",".join(extra_fields)}'
            )

        result = {}
        for setting_name in self.settings:
            result[setting_name] = self.validate_setting(config, setting_name)

        return result


class SpecValidatedConfig(CompositeConfig):
    """An all-in-one class that allows to specify settings and validate values;
    takes an iterable of configs as its source of values.

    Also defines __getattr__ so settings can be accessed as properties
    instead of indexing.

    :param subconfigs: A number of source configs, from lowest priority to highest.
    :param validate: Whether to validate the config right after initialization.

    Example:

    .. code-block:: python

        class ExampleToolConfig(cfglib.SpecValidatedConfig):
            message = cfglib.StringSetting(default='Hello!')
            config_file = cfglib.StringSetting(default=None)
    """

    allow_extra = False
    """Whether source configs can include extra keys not from the spec."""

    SPEC: ExtOptional[ConfigSpec] = None
    """ConfigSpec of this config."""

    def __init_subclass__(cls, **kwargs):  # pylint: disable=unused-argument
        super().__init_subclass__(**kwargs)

        if getattr(cls, 'SPEC', None) is not None:
            return

        settings = []
        for name, setting in cls.__dict__.items():
            if not isinstance(setting, Setting):
                continue

            if name != setting.name:
                raise ValueError(
                    f'Mismatch between setting\'s attribute {name}'
                    f' and its name {setting.name}'
                )

            settings.append(setting)

        spec = ConfigSpec(settings, allow_extra=cls.allow_extra)
        cls.SPEC = spec

    def __init__(
        self,
        subconfigs: Union[Mapping, Iterable[Mapping]],
        validate=True,
    ):
        super().__init__(to_cfg_list(subconfigs))

        # Store a second composite config that can be passed to spec validation
        # as a plain ordinary config
        self._composite_config = CompositeConfig([])
        self._composite_config.subconfigs = self.subconfigs

        if validate:
            self.validate()

    def validate(self):
        """Revalidate this config according to the spec."""
        self.SPEC.validate_config(self._composite_config)

    def __getitem__(self, item):
        value = self.SPEC.validate_setting(self._composite_config, item)

        if value is MISSING:
            raise KeyError(f'Key {item} not found')

        return value

    def __iter__(self):
        return (key for key in self.SPEC.settings if self.SPEC.validate_setting(self, key))

    def __len__(self):
        return len(list(self.__iter__()))

    def __getattr__(self, item):
        try:
            return self.__getitem__(item)
        except KeyError as exc:
            raise AttributeError(*exc.args)

    def __repr__(self):
        snapshot = self.snapshot()
        return f'<{self.__class__.__name__} {snapshot}>'
